package Event::RPC;

$VERSION = "0.80";

__END__

=head1 NAME

Event::RPC - Event based transparant Client/Server RPC mechanism

=head1 SYNOPSIS

  #-- Server Code
  use Event::RPC::Server;
  $server = Event::RPC::Server->new (
      name    => "test daemon",
      port    => 5555,
      classes => { "Event::RPC::Test" => { ... } },
  );
  $sever->start;
  
  #-- Client Code
  use Event::RPC::Client;
  my $client = Event::RPC::Client->connect (
      server   => "localhost",
      port     => 5555,
  );
  #-- Call methods of Event::RPC::Test on the server

=head1 ABSTRACT

Event::RPC supports you in developing Event based Client/Server applications with transparent object/method access from the client to the server.

=head1 DESCRIPTION

Event::RPC consists of a server and a client library. The server exports a list of modules and methods, which are allowed to be called over the network. More specific it acts as a proxy for objects created on the server side (on demand of the connected clients) which handles client side methods calls with transport of method arguments and return values.

The server side object proxy handles refcounting and destruction of objects created by clients properly. Objects as method parameters and return values are handled as well.

For the client the whole thing is totally transparent - once connected to the server it doesn't know whether it calls methods on local or remote objects.

For details please refer to the man pages of Event::RPC::Server and Event::RPC::Client.

=head1 EXAMPLE

  Server:
  ==================================================

  use strict;
  use Event::RPC::Server;

  main: {
      #-- Create a Server instance and declare the
      #-- exported interface
      my $server = Event::RPC::Server->new (
        name    => "test daemon",
        port    => 5555,
        classes => {
          'Event::RPC::Test' => {
            new       => '_constructor', # Class constructor
            set_data  => 1,              # and 'normal' methods...
            get_data  => 1,
            hello     => 1,
          },
        },
      );

      #-- Start the server resp. the Event loop.
      $server->start;
  }

  #-- A simple test class

  package Event::RPC::Test;

  sub get_data  { shift->{data}         }
  sub set_data  { shift->{data} = $_[1] }

  sub new {
      my $class = shift;
      my %par = @_;
      my ($data) = $par{'data'};

      my $self = bless { data => $data };

      return $self;
  }

  sub hello {
      my $self = shift;
      return "I have this data: '".$self->get_data."'";
  }

  Client:
  ==================================================

  use strict;

  use Event::RPC::Client;

  main: {
      #-- This connects to the server, requests the exported
      #-- interfaces and establishes proxy methods
      #-- in the correspondent packages.
      my $client = Event::RPC::Client->connect (
        server   => "localhost",
        port     => 5555,
        error_cb => sub { print "An RPC error occured\n"; exit },
      );

      #-- From now on the call to Event::RPC::Test->new is handled
      #-- transparently by Event::RPC::Client
      my $object = Event::RPC::Test->new (
              data => "some test data" x 5
      );

      #-- and method calls as well...
      print "hello=".$object->hello,"\n";
      $object->set_data ("changed data");
      print "data=".$object->get_data."\n";

      #-- disconnection is handled by the destructor of $client 
  }

=head1 AUTHORS

  Jrn Reder <joern at zyn dot de>

=head1 COPYRIGHT AND LICENSE

Copyright 2002-2005 by Jrn Reder.

This library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2.1 of the
License, or (at your option) any later version.

This library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307
USA.

=cut
