# $Id: Server.pm,v 1.15 2004/04/13 20:47:01 joern Exp $

#-----------------------------------------------------------------------
# Copyright (C) 2001-2003 Jrn Reder <joern AT zyn.de>.
# All Rights Reserved. See file COPYRIGHT for details.
# 
# This module is part of Event::RPC, which is free software; you can
# redistribute it and/or modify it under the same terms as Perl itself.
#-----------------------------------------------------------------------

package Event::RPC::Server;

use Event::RPC::Message;

use Carp;
use strict;
use IO::Socket::INET;
use Sys::Hostname;

use Event;

sub get_port			{ shift->{port}				}
sub get_name			{ shift->{name}				}
sub get_classes			{ shift->{classes}			}
sub get_loaded_classes		{ shift->{loaded_classes}		}
sub get_clients_connected	{ shift->{clients_connected}		}
sub get_log_clients_connected	{ shift->{log_clients_connected}	}
sub get_logging_clients		{ shift->{logging_clients}		}
sub get_logger			{ shift->{logger}			}
sub get_objects			{ shift->{objects}			}

sub set_port			{ shift->{port}			= $_[1]	}
sub set_name			{ shift->{name}			= $_[1]	}
sub set_classes			{ shift->{classes}		= $_[1]	}
sub set_loaded_classes		{ shift->{loaded_classes}	= $_[1]	}
sub set_clients_connected	{ shift->{clients_connected}	= $_[1]	}
sub set_log_clients_connected	{ shift->{log_clients_connected}= $_[1]	}
sub set_logging_clients		{ shift->{logging_clients}	= $_[1]	}
sub set_logger			{ shift->{logger}		= $_[1]	}
sub set_objects			{ shift->{objects}		= $_[1]	}

my $INSTANCE;
sub instance { $INSTANCE }

sub new {
	my $class = shift;
	my %par = @_;
	my  ($port, $classes, $name, $logger) =
	@par{'port','classes','name','logger'};
	
	my $self = bless {
		port			=> $port,
		name			=> $name,
		logger			=> $logger,
		classes			=> $classes,
		loaded_classes		=> {},
		objects			=> {},
		logging_clients		=> {},
		clients_connected	=> 0,
	}, $class;

	$INSTANCE = $self;

	$self->log ($self->get_name." started");

	return $self;
}

sub start {
	my $self = shift;

	my $port = $self->get_port;

	# setup rpc listener
	my $rpc_socket = IO::Socket::INET->new (
		Listen    => 5,
		LocalPort => $port,
		Proto     => 'tcp',
	) or die "can't start RPC listener: $!";

	Event->io (
		fd        => $rpc_socket,
		poll      => 'r',
		cb        => sub { $self->accept_new_client($rpc_socket) },
		desc      => "rpc listener port $port",
		reentrant => 0,
	);

	$self->log ("Started rpc listener on port $port");

	# setup log listener
	my $log_socket = IO::Socket::INET->new (
		Listen    => 5,
		LocalPort => $port + 1,
		Proto     => 'tcp',
	) or die "can't start log listener: $!";

	Event->io (
		fd        => $log_socket,
		poll      => 'r',
		cb        => sub { $self->accept_new_log_client($log_socket) },
		desc      => "log listener port ".($port+1),
		reentrant => 0,
	);

	$self->log ("Started log listener on port ".($port+1));

	Event::loop();

	$self->log ("Server stopped");

	1;
}

sub accept_new_client {
	my $self = shift;
	my ($rpc_socket) = @_;

	my $client_socket = $rpc_socket->accept or return;

	Event::RPC::Server::Client->new (
		sock   => $client_socket,
		server => $self,
	);

	$self->set_clients_connected ( 1 + $self->get_clients_connected );

	1;
}

sub accept_new_log_client {
	my $self = shift;
	my ($log_socket) = @_;
	
	my $client_socket = $log_socket->accept or return;

	my $log_client = Event::RPC::Server::LogClient->new (
		sock   => $client_socket,
		server => $self,
	);

	$self->set_log_clients_connected ( 1 + $self->get_log_clients_connected );
	$self->get_logging_clients->{$log_client->get_cid} = $log_client;

	$self->get_logger->add_fh($client_socket);

	$self->log(2, "New log client connected");

	1;
}

sub load_class {
	my $self = shift;
	my %par = @_;
	my ($class) = $par{'class'};

	my $client = Event::RPC::Server::Client->new (
		server => $self,
	);
	
	$client->load_class ( class => $class );
	
	$class;
}

sub log {
	my $self = shift;
	my $logger = $self->get_logger;
	return unless $logger;
	$logger->log(@_);
	1;
}

sub remove_object {
	my $self = shift;
	my ($object) = @_;
	
	my $objects = $self->get_objects;

	die "Object $object not registed" if not $objects->{"$object"};

	delete $objects->{"$object"};
	
	$self->log(5, "Object '$object' removed");

	1;
}

sub register_object {
	my $self = shift;
	my ($object, $class) = @_;
	
	my $objects = $self->get_objects;

	my $refcount;
	if ( $objects->{"$object"} ) {
		$refcount = ++$objects->{"$object"}->{refcount};
	} else {
		$refcount = 1;
		$objects->{"$object"} = {
			object   => $object,
			class    => $class,
			refcount => 1,
		};
	}
	
	$self->log(5, "Object '$object' registered. Refcount=$refcount");
	
	1;
}

sub deregister_object {
	my $self = shift;
	my ($object) = @_;
	
	my $objects = $self->get_objects;

	die "Object $object not registered" if not $objects->{"$object"};
	
	my $refcount = --$objects->{"$object"}->{refcount};

	$self->log(5, "Object '$object' deregistered. Refcount=$refcount");

	$self->remove_object($object) if $refcount == 0;
		
	1;
}

sub print_object_register {
	my $self = shift;
	
	print "-"x70,"\n";

	my $objects = $self->get_objects;
	foreach my $oid ( sort keys %{$objects} ) {
		print "$oid\t$objects->{$oid}->{refcount}\n";
	}
	
	1;
}

package Event::RPC::Server::Client;

use Carp;
use Socket;

my $CONNECTION_ID;

sub get_cid			{ shift->{cid}				}
sub get_sock			{ shift->{sock}				}
sub get_server			{ shift->{server}			}
sub get_watcher			{ shift->{watcher}			}

sub get_classes			{ shift->{server}->{classes}		}
sub get_loaded_classes		{ shift->{server}->{loaded_classes}	}
sub get_objects			{ shift->{server}->{objects}		}
sub get_client_objects		{ shift->{client_objects}		}

sub set_watcher			{ shift->{watcher}		= $_[1]	}

sub get_message			{ shift->{message}			}
sub set_message			{ shift->{message}		= $_[1]	}

sub new {
	my $class = shift;
	my %par = @_;
	my  ($sock, $server) =
	@par{'sock','server'};

	my $cid = ++$CONNECTION_ID;
	
	my $self = bless {
		cid     => $cid,
		sock    => $sock,
		server  => $server,
		watcher => undef,
		message => undef,
		client_objects => {},
	}, $class;

	if ( $sock ) {
		$self->log (2,
			"Got new RPC connection. Connection ID is $cid"
		);
		$self->{watcher} = Event->io (
			fd   => $sock,
			poll => 'r',
			cb   => [ $self, "input" ],
			desc => "rpc client cid=$cid",
			reentrant => 0,
		);
	}
	
	return $self;
}

sub disconnect {
	my $self = shift;

	close $self->get_sock;
	$self->get_watcher->cancel;
	$self->set_watcher(undef);

	my $server = $self->get_server;

	$server->set_clients_connected ( $self->get_server->get_clients_connected - 1 );

	foreach my $oid ( keys %{$self->get_client_objects} ) {
		$server->deregister_object($oid);
	}

	$self->log(2, "Client disconnected");

	1;
}

sub log {
	my $self = shift;

	my ($level, $msg);
	if ( @_ == 2 ) {
		($level, $msg) = @_;
	} else {
		($msg) = @_;
		$level = 1;
	}

	$msg = "cid=".$self->get_cid.": $msg";
	
	return $self->get_server->log ($level, $msg);
}

sub input {
	my $self = shift;
	my ($e) = @_;

	my $message = $self->get_message;

	if ( not $message ) {
		$message = Event::RPC::Message->new ( sock => $self->get_sock );
		$self->set_message($message);
	}

	my $request = eval { $message->read };
	my $error = $@;

	return if not defined $request and not $error;

	$self->set_message(undef);

	return $self->disconnect
		if $request eq "DISCONNECT\n" or
		   $error =~ /DISCONNECTED/;

	$self->log (6, "Length of input request = ".length($request));

	my ($cmd, $rc);
	$cmd = $request->{cmd} if not $error;
	
	if ( $error ) {
		$self->log ("Unexpected error on incoming RPC call: $@");
		$rc = {
			ok  => 0,
			msg => "Unexpected error on incoming RPC call: $@",
		};
	} elsif ( $cmd eq 'new' ) {
		$rc = $self->create_new_object (
			request => $request,
		);

	} elsif ( $cmd eq 'exec' ) {
		$rc = $self->execute_object_method (
			request => $request,
		);

	} elsif ( $cmd eq 'classes_list' ) {
		$rc = $self->get_classes_list (
			request => $request,
		);

	} elsif ( $cmd eq 'class_info' ) {
		$rc = $self->get_class_info (
			request => $request,
		);

	} elsif ( $cmd eq 'client_destroy' ) {
		$rc = $self->object_destroyed_on_client (
			request => $request,
		);

	} else {
		$self->log ("Unknown request command '$cmd'");
		$rc = {
			ok  => 0,
			msg => "Unknown request command '$cmd'",
		};
	}

	$message->write($rc) and return;

	Event->io (
		fd        => $self->get_sock,
		poll      => 'w',
		cb        => sub { $_[0]->w->cancel if $message->write },
		reentrant => 0,
	);

	1;
}

sub create_new_object {
	my $self = shift;
	my %par = @_;
	my ($request) = $par{'request'};

	# Let's create a new object
	my $class_method = $request->{method};
	my $class = $class_method;
	$class =~ s/::[^:]+$//;
	$class_method =~ s/^.*:://;

	# check if access to this class/method is allowed
	if ( not defined $self->get_classes->{$class}->{$class_method} or
	     $self->get_classes->{$class}->{$class_method} ne '_constructor' ) {
		$self->log ("Illegal constructor access to $class->$class_method");
		return {
			ok  => 0,
			msg => "Illegal constructor access to $class->$class_method"
		};

	}
	
	# load the class if not done yet
	$self->load_class ( class => $class );

	# resolve object params
	$self->resolve_object_params ( params => $request->{params} );

	# ok, the class is there, let's execute the method
	my $object = eval {
		$class->$class_method (@{$request->{params}})
	};

	# report error
	if ( $@ ) {
		$self->log ("Error: can't create object ".
			    "($class->$class_method): $@");
		return {
			ok  => 0,
			msg => $@,
		};
	}

	# register object
	$self->get_server->register_object ($object, $class);
	$self->get_client_objects->{"$object"} = 1;

	# log and return
	$self->log (5,
		"Created new object $class->$class_method with oid '$object'",
	);

	return {
		ok  => 1,
		oid => "$object",
	};
}

sub load_class {
	my $self = shift;
	my %par = @_;
	my ($class) = $par{'class'};
	
	my $mtime;
	my $load_class_info = $self->get_loaded_classes->{$class};

	if ( not $load_class_info or
	     ( $mtime = (stat($load_class_info->{filename}))[9])
		> $load_class_info->{mtime} ) {
	
		if ( not $load_class_info->{filename} ) {
			my $filename;
			my $rel_filename = $class;
			$rel_filename =~ s!::!/!g;
			$rel_filename .= ".pm";

			foreach my $dir ( @INC ) {
				$filename = "$dir/$rel_filename", last
					if -f "$dir/$rel_filename";
			}

			croak "File for class '$class' not found"
				if not $filename;
			
			$load_class_info->{filename} = $filename;
		}
	
		$self->log (3, "Class '$class' ($load_class_info->{filename}) changed on disk. Reloading...")
			if $mtime > $load_class_info->{mtime};

		do $load_class_info->{filename};

		if ( $@ ) {
			$self->log ("Can't load class '$class': $@");
			$load_class_info->{mtime} = 0;

			return {
				ok  => 0,
				msg => "Can't load class $class: $@",
			};

		} else {
			$self->log (3, "Class '$class' successfully loaded");
			$load_class_info->{mtime} = time;
		}
	}
	
	$self->log (3, "filename=".$load_class_info->{filename}.
		    ", mtime=".$load_class_info->{mtime} );

	$self->get_loaded_classes->{$class} ||= $load_class_info;

	1;
}

sub execute_object_method {
	my $self = shift;
	my %par = @_;
	my ($request) = $par{'request'};

	# Method call of an existent object
	my $oid = $request->{oid};
	my $object_entry = $self->get_objects->{$oid};
	my $method = $request->{method};

	if ( not defined $object_entry ) {
		# object does not exists
		$self->log ("Illegal access to unknown object with oid=$oid");
		return {
			ok  => 0,
			msg => "Illegal access to unknown object with oid=$oid"
		};

	}
	
	my $class = $object_entry->{class};
	if ( not defined $self->get_classes->{$class}->{$method} ) {
		# illegal access to this method
		$self->log ("Illegal access to $class->$method");
		return {
			ok  => 0,
			msg => "Illegal access to $class->$method"
		};

	}
	
	# (re)load the class if not done yet
	$self->load_class ( class => $class );

	# resolve object params
	$self->resolve_object_params ( params => $request->{params} );

	# ok, try executing the method
	my @rc = eval {
		$object_entry->{object}->$method (@{$request->{params}})
	};

	# report error
	if ( $@ ) {
		$self->log ("Error: can't call '$method' of object ".
			    "with oid=$oid: $@");
		return {
			ok  => 0,
			msg => $@,
		};
	}
	
	# log
	$self->log (4, "Called method '$method' of object ".
		       "with oid=$oid");

	# check if objects are returned by this method
	# and register them in our internal object table
	# (if not already done yet)
	my $key;
	foreach my $rc ( @rc ) {
		if ( ref ($rc) and ref ($rc) !~ /ARRAY|HASH|SCALAR/ ) {
			# returns a single object
			$self->log (4, "Method returns object: $rc");
			$key = "$rc";
			$self->get_client_objects->{$key} = 1;
			$self->get_server->register_object($rc, ref $rc);
			$rc = $key;

		} elsif ( ref $rc eq 'ARRAY' ) {
			# possibly returns a list of objects
			# make a copy, otherwise the original object references
			# will be overwritten
			my @val = @{$rc};
			$rc = \@val;
			foreach my $val ( @val ) {
				if ( ref ($val) and ref ($val) !~ /ARRAY|HASH|SCALAR/ ) {
					$self->log (4, "Method returns object lref: $val");
					$key = "$val";
					$self->get_client_objects->{$key} = 1;
					$self->get_server->register_object($val, ref $val);
					$val = $key;
				}
			}
		} elsif ( ref $rc eq 'HASH' ) {
			# possibly returns a hash of objects
			# make a copy, otherwise the original object references
			# will be overwritten
			my %val = %{$rc};
			$rc = \%val;
			foreach my $val ( values %val ) {
				if ( ref ($val) and ref ($val) !~ /ARRAY|HASH|SCALAR/ ) {
					$self->log (4, "Method returns object href: $val");
					$key = "$val";
					$self->get_client_objects->{$key} = 1;
					$self->get_server->register_object($val, ref $val);
					$val = $key;
				}
			}
		}
	}

	# return rc
	return {
		ok => 1,
		rc => \@rc,
	};
}

sub object_destroyed_on_client {
	my $self = shift;
	my %par = @_;
	my ($request) = $par{'request'};

	$self->log(5, "Object with oid=$request->{oid} destroyed on client");

	delete $self->get_client_objects->{$request->{oid}};
	$self->get_server->deregister_object($request->{oid});

	return {
		ok => 1
	};
}

sub get_classes_list {
	my $self = shift;
	my %par = @_;
	my ($request) = $par{'request'};

	my @classes = keys %{$self->get_classes};
	
	return {
		ok      => 1,
		classes => \@classes,
	}
}

sub get_class_info {
	my $self = shift;
	my %par = @_;
	my ($request) = $par{'request'};

	my $class = $request->{class};
	
	if ( not defined $self->get_classes->{$class} ) {
		$self->log ("Unknown class '$class'");
		return {
			ok  => 0,
			msg => "Unknown class '$class'"
		};
	}
	
	$self->log (4, "Class info for '$class' requested");

	return {
		ok           => 1,
		methods      => $self->get_classes->{$class},
	};
}

sub resolve_object_params {
	my $self = shift;
	my %par = @_;
	my ($params) = $par{'params'};
	
	my $key;
	foreach my $par ( @{$params} ) {
		if ( defined $self->get_classes->{ref($par)} ) {
			$key = ${$par};
			$key = "$key";
			croak "unknown object with key '$key'"
				if not defined $self->get_objects->{$key};
			$par = $self->get_objects->{$key}->{object};
		}
	}
	
	1;
}


package Event::RPC::Server::LogClient;

use Carp;
use Socket;

my $LOG_CONNECTION_ID;

sub get_cid			{ shift->{cid}				}
sub get_sock			{ shift->{sock}				}
sub get_server			{ shift->{server}			}
sub get_watcher			{ shift->{watcher}			}

sub new {
	my $class = shift;
	my %par = @_;
	my  ($sock, $server) =
	@par{'sock','server'};

	my $cid = ++$LOG_CONNECTION_ID;
	
	my $self = bless {
		cid     => $cid,
		sock    => $sock,
		server  => $server,
		watcher => undef,
	}, $class;

	$self->{watcher} = Event->io (
		fd   => $sock,
		poll => 're',
		cb   => [ $self, 'input' ],
		desc => "log reader $cid",
		reentrant => 0,
	);

	$self->get_server->log (2,
		"Got new logger connection. Connection ID is $cid"
	);

	return $self;
}

sub disconnect {
	my $self = shift;

	my $sock = $self->get_sock;
	$self->get_server->get_logger->remove_fh($sock);
	$self->get_watcher->cancel;
	close $sock;

	$self->get_server->set_log_clients_connected ( $self->get_server->get_log_clients_connected - 1 );
	delete $self->get_server->get_logging_clients->{$self->get_cid};
	$self->get_server->log(2, "Log client disconnected");

	1;
}

sub input {
	my $self = shift;

	my $sock = $self->get_sock;
	$self->disconnect if eof($sock);
	<$sock>;
	
	1;
}

1;
